/* Easy Machine: utility header.
   This is a derivative work based on Jitter, which is released under
   the GNU General Public License, version 3 or later. */

#ifndef EASY_MACHINE_UTILITY_H_
#define EASY_MACHINE_UTILITY_H_

#include <stdarg.h>
#include <stdlib.h>


/* Fatal errors.
 * ************************************************************************** */

#define easy_machine_fatal(...)                                     \
  do                                                                \
    {                                                               \
      /* Ignore any failures in printing a fatal error message. */  \
      printf ("FATAL ERROR: " __VA_ARGS__);                         \
      printf ("\n");                                                \
      exit (EXIT_FAILURE);                                          \
    }                                                               \
  while (false)

#define easy_machine_exception(...)                                 \
  do                                                                \
    {                                                               \
      printf ("\nEXCEPTION: ");                                     \
      printf (__VA_ARGS__);                                         \
      printf ("\n");                                                \
      printf ("The kernel killed your process.  Goodbye.\n");       \
      exit (EXIT_FAILURE);                                          \
    }                                                               \
  while (false)




/* Embedding C code into literal strings.
 * ************************************************************************** */

/* Return a freshly malloced copy of the pointed string. */
char *
easy_machine_string_clone (const char *s)
  __attribute__ ((nonnull (1), returns_nonnull));




/* Number parsing
 * ************************************************************************** */

long long
easy_machine_scan_literal (const char *s);



/* Bitwise tricks.
 * ************************************************************************** */

/* Given a possibly signed integer x and a power of two p, expand to an
   expression evaluating to the minimum integer greater than or equal to x which
   is a multiple of p.

   This assumes that p is a power of two and may evaluate both arguments
   multiple times.  However, if both arguments are constant expression, the
   expansion is also a constant expression.

   I learned this technique from from Hacker's Delight, §3-1.
   Let y be the smallest multiple of p greater than or equal to x, where p
   is a power of two.  Then:
     y = (x + p - 1) & - p     [1]
   and also
     y = x + (- x & (p - 1))   [2]
   I am currently using [1].  [2] can be useful in other contexts, because
   the second addend expresses a displacement from x to y. */
#define JITTER_NEXT_MULTIPLE_OF_POWER_OF_TWO(x, p)  \
  (((x) + ((p) - 1)) & ~ ((p) - 1))




/* Embedding C code into literal strings.
 * ************************************************************************** */

/* A CPP hack: this is a way of stringizing arbitrary C code (not including,
   unfortunately, CPP directives), as tokenized by CPP.  
   In theory this could be non-portable.  In practice it works fine, even if the
   C++ (and, I guess C as well) standard only requires compilers to accept a
   limited-length logical line.  This limit is 65536 characters, which should be
   enough if long literal code is split across different macro calls once in a
   while.  Anyway, this is dumb.  GCC, like any other well-designed compiler,
   has no such artificial limit. */
#define CCODE(...)  \
  #__VA_ARGS__

#endif // #ifndef EASY_MACHINE_UTILITY_H_
